# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bit field structure member definition."""

# pylint: disable=too-many-arguments

from dataclasses import dataclass
from enum import IntEnum, EnumMeta
from typing import Any, Dict, Optional, Tuple

from .._code_injector import CodeInjector
from .._getbytes import getbytes
from .._typing import ByteOrderHint
from ..bitfields import BitFields
from ..dump import Record
from ..enum import EnumX
from ..flag import FlagX
from ..exceptions import InsufficientMemoryError
from ..none import NoneX
from ._member import Member


@dataclass
class BitFieldsStore:

    """Temporary information store for generating implementation code."""

    is_first: bool = False
    is_last: bool = False
    group_name: str = ""
    nbytes: Optional[int] = None
    byteorder: str = "little"


class BitFieldMember(Member):

    """Bit field structure member definition.

    :param doc: accessor documentation string
    :param size: bit field width (in bits)
    :param lsb: least significant bit position
    :param typ: field type (e.g. ``IntEnum`` subclass), default uses type annotation
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param compute: initializer defaults to compute based on another member
    :param nbytes: bytes occupied by this bit field and ones that follow (0 => calculate)
    :param argrepr: format to represent member argument in structure repr

    """

    aux_store: BitFieldsStore

    _group_name: str
    _typ: Any

    _attributes_to_copy: Tuple[str, ...] = Member._attributes_to_copy + (
        "_typ",
        "_size",
        "_lsb",
        "_signed",
        "_group_name",
    )

    def __init__(
        self,
        doc: str,
        size: int,
        lsb: Optional[int],
        typ: type,
        signed: bool,
        default: Any,
        ignore: bool,
        readonly: bool,
        compute: bool,
        nbytes: int,
        argrepr: Optional[str],
    ) -> None:
        if lsb is not None:
            lsb = int(lsb)
            if lsb < 0:
                raise TypeError(
                    "bit field position must be greater than or equal to zero"
                )

        if signed and size < 2:
            raise ValueError("'size' must be 2 or greater for signed bit field")

        if size < 1:
            raise ValueError("'size' must be 1 or greater for unsigned bit field")

        super().__init__(
            doc,
            default,
            ignore,
            readonly,
            compute,
            argrepr,
            fmt=NoneX("none"),
            fmt_arg=None,
        )

        self._set_typ(typ)
        self._size = size
        self._lsb = lsb
        self._nbytes = nbytes
        self._signed = signed
        self._group_name = ""

    @property
    def lsb(self) -> int:
        """Bit field bit position."""
        return -1 if self._lsb is None else self._lsb

    @property
    def nbytes(self) -> int:
        """Number of bytes allocated for this and following bit fields.

        Return zero for bit fields included in previous bit field allocation.

        """
        if self._nbytes is None:  # pragma: no cover
            raise RuntimeError("internal error")
        return self._nbytes

    @property
    def signed(self) -> bool:
        """Bit field is a signed integer."""
        return self._signed

    @property
    def size(self) -> int:
        """Bit field width (in bits)."""
        return self._size

    @property
    def typ(self) -> type:
        """Field type (e.g. ``IntEnum`` subclass), default uses type annotation."""
        if self._typ is None:  # pragma: no cover
            raise RuntimeError("internal error")

        return self._typ

    def _set_typ(self, typ: type):
        try:
            is_int_subclass = issubclass(typ, (int, BitFields))

        except TypeError:
            is_int_subclass = False  # must not be class

        if isinstance(typ, (FlagX, EnumX)) or is_int_subclass:
            pass

        else:
            raise TypeError("bit field type must be int-like")

        self._typ = typ

    @property
    def argrepr(self) -> str:
        """Format string to represent member argument in structure repr."""
        if isinstance(self._argrepr, str):
            repr_format = self._argrepr

        elif isinstance(self._typ, EnumMeta):
            repr_format = f"{self.name}={{str(self.{self.name})}}"

        else:
            repr_format = f"{self.name}={{self.{self.name}!r}}"

        return repr_format

    @classmethod
    def organize(cls, members: Dict[str, Member], byteorder: str, fieldorder: str):
        """Organize bit field members into groups."""
        collection = []
        group = {}

        for name, member in members.items():
            if isinstance(member, BitFieldMember) and not member.nbytes:
                group[name] = member
                continue

            if group:
                collection.append(group)
                group = {}

            if isinstance(member, BitFieldMember):
                group[name] = member

        if group:
            collection.append(group)

        for i, group in enumerate(collection):
            nbytes = cls.position_and_size(f"fields_{i}", group, fieldorder)
            for member in group.values():
                member.aux_store = BitFieldsStore(
                    byteorder=byteorder, nbytes=nbytes, group_name=f"bitfields_{i}"
                )
            list(group.values())[0].aux_store.is_first = True
            list(group.values())[-1].aux_store.is_last = True

    @staticmethod
    def position_and_size(
        group_name: str, members: Dict[str, "BitFieldMember"], fieldorder: str
    ) -> int:
        """Fill in any unassigned positions, determine overall size."""
        # fill in missing positions (and assign group name)
        lsb = 0
        field_member_properties = (
            members.values()
            if fieldorder == "least_to_most"
            else reversed(list(members.values()))
        )
        for field_member in field_member_properties:
            # pylint: disable=protected-access
            field_member._group_name = group_name
            if field_member._lsb is None:
                field_member._lsb = lsb
            lsb = field_member.lsb + field_member.size

        # check for overlap
        claimed_bits: Dict[int, str] = {}
        for name, member in members.items():
            for i in range(member.lsb, member.lsb + member.size):
                if i in claimed_bits:
                    raise TypeError(
                        f"bit field {name!r} overlaps with {claimed_bits[i]!r}"
                    )

                claimed_bits[i] = name

        # default/validate 'nbytes'
        numbits = max(member.lsb + member.size for member in members.values())
        nbytes_needed = (numbits + 7) // 8

        first_bitfield_name, first_bitfield = list(members.items())[0]
        nbytes = first_bitfield.nbytes

        if not nbytes:
            nbytes = first_bitfield._nbytes = nbytes_needed

        elif nbytes_needed > nbytes:
            raise TypeError(
                f"'nbytes' must be at least {nbytes_needed} for bitfield "
                f"{first_bitfield_name!r} (to have sufficient room for it "
                f"and those that follow)"
            )

        return nbytes

    @staticmethod
    def __unpack__(
        buffer: bytes,
        offset: int,
        dump: Optional[Record],
        byteorder: ByteOrderHint,
        nbytes: int,
        signed: bool,
    ) -> Tuple[int, int]:
        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError(
                    "too few bytes to unpack, retry with dump to follow"
                )

            value = int.from_bytes(buffer[offset:end], byteorder, signed=signed)

            return value, end

        chunk, offset = getbytes(buffer, offset, dump, nbytes)

        value = int.from_bytes(chunk, byteorder, signed=signed)

        dump.value = value

        return value, offset

    def add_name_index(
        self,
        name: str,
        index: int,
        code_injector: CodeInjector,
    ) -> int:
        """Assign name, index number, and prepare for code generation."""
        # pylint: disable=too-many-statements, too-many-branches, too-many-locals
        index = super().add_name_index(name, index, code_injector)

        store = self.temp_store
        aux = self.aux_store

        # Erase normal member packing lines (from super)
        store.unpack = []
        store.unpack_and_dump = []
        store.pack = []
        store.pack_and_dump = []

        lo_mask = (1 << self.size) - 1
        sign_bit = 1 << (self.size - 1)

        cls_name = self.typ.__name__
        try:
            is_same = code_injector.evaluate(cls_name) is self.typ
        except NameError:  # pragma: no cover
            is_same = False
        cls_name = cls_name if is_same else f"cls.{self.name}.typ"

        dump_name = f"{aux.group_name}_dump"
        dump_group = [f"{dump_name} = dump.add_record()"]

        if issubclass(self.typ, BitFields):
            dump_member = [
                # line 1
                f"dump_m_{self.name} = dump.add_record("
                f'access="{name}", value=int(m_{self.name}), fmt={cls_name})',
                # line 2
                f"{cls_name}.__add_bitfields_to_dump__("
                f"m_{self.name}, dump_m_{self.name}, bitoffset={self.lsb})",
            ]
        else:
            dump_member = [
                f"dump.add_record("
                f'access="{name}", bits={self.lsb, self.size}, '
                f"value=m_{self.name}, fmt={cls_name})"
            ]

        if aux.is_first:
            store.unpack += [
                f"{aux.group_name}, offset = "
                f"cls.{name}.__unpack__(buffer, offset, dump, "
                f'"{aux.byteorder}", nbytes={aux.nbytes}, signed={self.signed})',
                "",
            ]

            store.unpack_and_dump += dump_group + [
                f"{aux.group_name}, offset = "
                f"cls.{name}.__unpack__(buffer, offset, {dump_name}, "
                f'"{aux.byteorder}", nbytes={aux.nbytes}, signed={self.signed})',
                "",
            ]

        lines = [f"m_{name} = ({aux.group_name} >> {self.lsb}) & 0x{lo_mask:x}"]

        if self.signed:
            lines += [
                f"if m_{name} & 0x{sign_bit:x}:",
                f"    m_{name} -= 0x{1 << self.size:x}",
            ]

        if self.typ is not int:
            if issubclass(self.typ, BitFields):
                lines.append(f"m_{name} = {cls_name}.from_int(m_{name})")
            else:
                lines.append(f"m_{name} = {cls_name}(m_{name})")

        store.unpack += lines
        store.unpack_and_dump += lines + dump_member

        if aux.is_first:
            lines = [f"{aux.group_name} = 0", ""]
            store.pack += lines
            store.pack_and_dump += dump_group + lines

        assert self.lsb is not None, "internal error"
        hi_mask = ~(lo_mask << self.lsb)
        member_name = f"m_{self.name}"
        if self.typ is bool:
            # don't let an integer > 1 be masked off and become False
            member_name = f"int(bool({member_name}))"
        op2 = f"({member_name} & 0x{lo_mask:x})"
        if self.lsb:
            op2 = f"({op2} << {self.lsb})"
        lines = [f"{aux.group_name} |= ({aux.group_name} & {hi_mask}) | {op2}"]
        store.pack += lines
        store.pack_and_dump += dump_member + lines

        if aux.is_last:
            store.pack += [
                "",
                f"pieces.append({aux.group_name}.to_bytes("
                f'{aux.nbytes}, "{aux.byteorder}", signed={self.signed}))',
            ]
            store.pack_and_dump += [
                "",
                f"pieces.append({aux.group_name}.to_bytes("
                f'{aux.nbytes}, "{aux.byteorder}", signed={self.signed}))',
                f"{dump_name}.value = str({aux.group_name})",
                f"{dump_name}.memory = pieces[-1]",
            ]

        if self.signed:
            neg_limit = -(1 << (self.size - 1))
            maxvalue = f"{-(1 + neg_limit)}"
            minvalue = f"-{-neg_limit}"
        else:
            minvalue = "0"
            maxvalue = f"{(1 << self.size) - 1}"

        if self.fset is None and not self.readonly:
            # override lines put in by super()
            if issubclass(self.typ, IntEnum):
                store.setter = (
                    []
                )  # no need for range check, enum instantiation checks validity
            else:
                store.setter = [
                    f"if not {minvalue} <= value <= {maxvalue}:",
                    f'   raise ValueError("out of range, {minvalue} <= value <= {maxvalue}")',
                ]
            if self.typ is int:
                store.setter.append(f"self[{self.index}] = value")
            else:
                store.setter.append(f"self[{self.index}] = {cls_name}(value)")

        if not self.default_is_a_factory:
            store.init_begin += [
                f"if not {minvalue} <= {self.name} <= {maxvalue}:",
                f'   raise ValueError("{self.name!r} out of range, {minvalue} <= {self.name} <= {maxvalue}")',
            ]

        return index


def bitfield_member(
    doc: str = "",
    *,
    size: int,
    lsb: Optional[int] = None,
    typ: type = int,
    signed: bool = False,
    default: Any = None,
    ignore: bool = False,
    readonly: bool = False,
    compute: bool = False,
    nbytes: int = 0,
    argrepr: Optional[str] = None,
) -> Any:
    """Bit field structure member definition.

    :param doc: accessor documentation string
    :param size: bit field width (in bits)
    :param lsb: least significant bit position
    :param signed: interpret as signed integer
    :param typ: field type (e.g. ``IntEnum`` subclass)
    :param default: initializer default value
    :param ignore: ignore member during comparisons
    :param readonly: block setting member attribute
    :param compute: initializer defaults to compute based on another member
    :param nbytes: bytes occupied by this bit field and ones that follow (0 => calculate)
    :param argrepr: format to represent member argument in structure repr

    """
    return BitFieldMember(
        doc, size, lsb, typ, signed, default, ignore, readonly, compute, nbytes, argrepr
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Float to bytes and bytes to float transform."""

from numbers import Real
from struct import Struct
from typing import List, Optional, Tuple

from ._getbytes import getbytes
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform
from .view import NumberView


FORMAT_CODES = {2: "e", 4: "f", 8: "d"}
ENDIAN_CODES = {"big": ">", "little": "<"}


class FloatView(NumberView, Real):

    """Float type view."""

    def __float__(self):
        return NumberView.unpack(self)


class FloatX(Transform):

    """Float to bytes and bytes to float transform."""

    def __init__(self, name: str, nbytes: int, byteorder: str = "little") -> None:
        assert nbytes in [2, 4, 8]
        assert byteorder in {"big", "little"}

        super().__init__(name)

        self.__nbytes__ = nbytes

        struct = Struct(ENDIAN_CODES[byteorder] + FORMAT_CODES[nbytes])

        self.__byteorder__ = byteorder
        self.__p__ = struct.pack
        self.__nu__ = nbytes, struct.unpack

    @property
    def byteorder(self) -> str:
        """Byte order ("little" or "big")."""
        return self.__byteorder__

    def __pack__(
        self, value: float, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        if dump is None:
            pieces.append(self.__p__(value))
        else:
            dump.value = value

            piece = self.__p__(value)

            dump.memory = piece

            pieces.append(piece)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[float, int]:
        nbytes, unpack = self.__nu__

        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            return unpack(buffer[offset:end])[0], end

        chunk, offset = getbytes(buffer, offset, dump, nbytes)

        value = dump.value = unpack(chunk)[0]

        return value, offset

    def __view__(self, buffer: bytearray, offset: int = 0) -> FloatView:
        """Create view of float in bytes buffer."""
        return FloatView(self, buffer, offset)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test BitField properties."""

import pytest

from baseline import Baseline

from plum.bitfields import BitFields, bitfield
from plum.conformance import wrap_message
from plum.enum import EnumX

from sample_bitfields import Register

register = EnumX("register", Register)


class Fields(BitFields, fieldorder="least_to_most"):

    defaulted_bitfield = bitfield("doc", size=3)

    kw_bitfield = bitfield(
        doc="kw doc",
        size=4,
        typ=bool,
        lsb=8,
        default=1,
        signed=True,
        ignore=True,
        readonly=True,
        argrepr="argrepr",
    )


class TestDefault:

    """Test with as many left to default as possible."""

    def test_argrepr(self):
        assert (
            Fields.defaulted_bitfield.argrepr
            == "defaulted_bitfield={self.defaulted_bitfield!r}"
        )

    def test_default(self):
        assert Fields.defaulted_bitfield.default is None

    def test_doc(self):
        assert Fields.defaulted_bitfield.doc == "doc"

    def test_ignore(self):
        assert Fields.defaulted_bitfield.ignore is False

    def test_lsb(self):
        assert Fields.defaulted_bitfield.lsb == 0

    def test_mask(self):
        assert Fields.defaulted_bitfield.mask == 7

    def test_minvalue(self):
        assert Fields.defaulted_bitfield.minvalue == 0

    def test_maxvalue(self):
        assert Fields.defaulted_bitfield.maxvalue == 7

    def test_name(self):
        assert Fields.defaulted_bitfield.name == "defaulted_bitfield"

    def test_readonly(self):
        assert Fields.defaulted_bitfield.readonly is False

    def test_signbit(self):
        assert Fields.defaulted_bitfield.signbit == 0

    def test_signed(self):
        assert Fields.defaulted_bitfield.signed is False

    def test_size(self):
        assert Fields.defaulted_bitfield.size == 3

    def test_type_hint(self):
        assert Fields.defaulted_bitfield.type_hint is int

    def test_typ(self):
        assert Fields.defaulted_bitfield.typ is int


class TestKeyword:

    """Test explicitly defined with keyword argument."""

    def test_argrepr(self):
        assert Fields.kw_bitfield.argrepr == "argrepr"

    def test_default(self):
        assert Fields.kw_bitfield.default == 1

    def test_doc(self):
        assert Fields.kw_bitfield.doc == "kw doc"

    def test_ignore(self):
        assert Fields.kw_bitfield.ignore is True

    def test_lsb(self):
        assert Fields.kw_bitfield.lsb == 8

    def test_mask(self):
        assert Fields.kw_bitfield.mask == 0xF

    def test_minvalue(self):
        assert Fields.kw_bitfield.minvalue == -8

    def test_maxvalue(self):
        assert Fields.kw_bitfield.maxvalue == 7

    def test_name(self):
        assert Fields.kw_bitfield.name == "kw_bitfield"

    def test_readonly(self):
        assert Fields.kw_bitfield.readonly is True

    def test_signbit(self):
        assert Fields.kw_bitfield.signbit == 0x8

    def test_signed(self):
        assert Fields.kw_bitfield.signed is True

    def test_size(self):
        assert Fields.kw_bitfield.size == 4

    def test_type_hint(self):
        assert Fields.kw_bitfield.type_hint is bool

    def test_typ(self):
        assert Fields.kw_bitfield.typ is bool


class TestEnum:

    """Test special handling for EnumX type hints."""

    def test_enumx_hint(self):
        bitfield_ = bitfield(size=2, typ=register)
        assert bitfield_.type_hint is Register


class TestReadOnly:
    def test_decorator(self):
        expected_message = Baseline(
            """
            'setter' not allowed on read-only bit field properties
            """
        )

        bitfield_ = bitfield(size=2, readonly=True)

        with pytest.raises(TypeError) as trap:

            @bitfield_.setter
            def bitfield_(self, value):
                # pylint: disable=unused-argument
                pass

        assert wrap_message(trap.value) == expected_message


class TestDeleter:
    def test_deleter_blocked(self):
        expected_message = Baseline(
            """
            bit field properties do not support 'deleter'
            """
        )

        bitfield_ = bitfield(size=2, readonly=True)

        with pytest.raises(TypeError) as trap:

            @bitfield_.deleter
            def bitfield_(self):
                # pylint: disable=unused-argument
                pass

        assert wrap_message(trap.value) == expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test miscellaneous array transform features."""

from baseline import Baseline
from plum.array import ArrayX
from plum.littleendian import uint8


class TestDumpWhenArrayEmpty:

    array_x = ArrayX(name="list", fmt=uint8)

    expected_dump = Baseline(
        """
        +--------+-------+-------+--------+
        | Offset | Value | Bytes | Format |
        +--------+-------+-------+--------+
        |        | []    |       | list   |
        +--------+-------+-------+--------+
        """
    )

    def test_pack(self):
        _, dump = self.array_x.pack_and_dump([])
        assert str(dump) == self.expected_dump

    def test_unpack(self):
        _, dump = self.array_x.unpack_and_dump(b"")
        assert str(dump) == self.expected_dump



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""AttrDict to bytes and bytes to AttrDict transform."""

from typing import Any, Dict, List, Optional, Tuple, Union

from .dump import Record
from .transform import Transform
from .exceptions import SizeError


class AttrDict(dict):

    """Dictionary with access to items as attributes."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def __repr__(self):
        params = [f"{key}={value!r}" for key, value in self.items()]
        return f"AttrDict({', '.join(params)})"


class AttrDictX(Transform):

    """AttrDict to bytes and bytes to AttrDict transform."""

    __nbytes__ = Optional[int]

    members: Dict[str, Any]

    def __init__(
        self,
        members: Union[Dict[str, Any], List[Tuple[str, Any]]],
        name: str = "AttrDict",
    ) -> None:
        super().__init__(name)
        self.members = dict(members)
        self.__nbytes__ = 0
        for transform in self.members.values():
            try:
                self.__nbytes__ += transform.nbytes
            except SizeError:
                self.__nbytes__ = None
                break

    def __unpack__(
        self,
        buffer: bytes,
        offset: int,
        dump: Optional[Record] = None,
    ) -> Tuple[AttrDict, int]:

        attrdict = AttrDict()

        if dump is None:
            for name, transform in self.members.items():
                value, offset = transform.__unpack__(buffer, offset, dump)
                attrdict[name] = value
        else:
            for name, transform in self.members.items():
                subdump = dump.add_record(access=name, fmt=transform)
                value, offset = transform.__unpack__(buffer, offset, subdump)
                attrdict[name] = value

        return attrdict, offset

    def __pack__(
        self, value, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        if dump is None:
            for name, transform in self.members.items():
                transform.__pack__(value[name], pieces, dump)

        else:
            for name, transform in self.members.items():
                subdump = dump.add_record(access=name, fmt=transform)
                subdump.value = member_value = value[name]
                transform.__pack__(member_value, pieces, subdump)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Bytes sequence incremental unpacker."""

from .dump import Dump
from .exceptions import ExcessMemoryError, ImplementationError, UnpackError
from .items import Format, items


class Buffer(bytes):

    """Bytes sequence incremental unpacker."""

    offset: int

    def __new__(cls, *args, **kwargs):
        instance = bytes.__new__(cls, *args, **kwargs)
        instance.offset = 0
        return instance

    def unpack(self, fmt: Format):
        """Unpack value from formatted bytes at current buffer offset.

        :raises: ``UnpackError`` if insufficient bytes or value error

        """
        try:
            # None -> dump
            value, self.offset = items.__unpack__(self, self.offset, None, fmt)
        except Exception as exc:
            # do it over to include dump in exception message
            self.unpack_and_dump(fmt)
            raise ImplementationError() from exc  # pragma: no cover

        return value

    def unpack_and_dump(self, fmt: Format):
        """Unpack value from bytes at current offset and produce a packed bytes summary.

        :raises: ``UnpackError`` if insufficient bytes or value error

        """
        offset = self.offset

        dump = Dump(offset=offset)

        try:
            # empty record added so so pack always called with a Record instance
            value, self.offset = items.__unpack__(self, offset, dump.add_record(), fmt)
        except Exception as exc:
            dump.trim_blank_record()
            raise UnpackError(dump, exc) from exc

        dump.trim_blank_record()

        return value, dump

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:  # don't check if already an exception
            extra_bytes = self[self.offset :]

            if extra_bytes:
                raise ExcessMemoryError(extra_bytes)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Common typing hints."""

import sys

if sys.version_info < (3, 8):
    ByteOrderHint = str
else:
    from typing import Literal, Union

    ByteOrderHint = Union[Literal["little"], Literal["big"]]



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Exception classes."""


class UnpackError(Exception):

    """Unpack error."""

    def __init__(self, dump=None, exception=None):
        message = (
            f"\n\n{dump}\n\n"
            f"{type(exception).__name__} occurred during unpack operation:"
            f"\n\n{exception}"
        )
        Exception.__init__(self, message)


class ExcessMemoryError(UnpackError):

    """Leftover bytes after unpack operation."""

    def __init__(self, extra_bytes=b""):
        # pylint: disable=non-parent-init-called,super-init-not-called
        message = f"{len(extra_bytes)} unconsumed bytes"
        Exception.__init__(self, message)
        self.extra_bytes = extra_bytes


class ImplementationError(Exception):

    """Unexpected implementation error."""

    def __init__(self, message=""):
        if not message:
            message = (
                "One of the plum types used in the pack/unpack operation "
                "contains an implementation error. The operation generated "
                "an exception when first performed without a dump (for efficiency). "
                "But when the operation was repeated with a dump (for a better "
                "exception message) the exception did not re-occur. Please report "
                "the inconsistent behavior to the type developer."
            )
        super().__init__(message)


class InsufficientMemoryError(UnpackError):

    """Too few bytes to unpack an item."""

    def __init__(self, message):
        # pylint: disable=non-parent-init-called,super-init-not-called
        Exception.__init__(self, message)


class PackError(Exception):

    """Pack operation error."""

    def __init__(self, message="", dump=None, exception=None):
        if dump:
            message += f"\n\n{dump}\n\n"

        if exception:
            message += f"{type(exception).__name__} occurred during pack operation:\n\n{exception}"

        Exception.__init__(self, message)


class SizeError(Exception):

    """Size varies from instance to instance."""



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test items transform with fmt=None and with dict of values."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.exceptions import PackError
from plum.items import items
from plum.littleendian import uint8, uint16


class Test:

    expected_dump = Baseline(
        """
        +--------+----------+-------+-------+--------+
        | Offset | Access   | Value | Bytes | Format |
        +--------+----------+-------+-------+--------+
        |        |          |       |       | items  |
        | 0      | ['mbr1'] | 1     | 01    | uint8  |
        | 1      | ['mbr2'] | 2     | 02 00 | uint16 |
        +--------+----------+-------+-------+--------+
        """
    )

    expected_bytes = bytes.fromhex("010200")

    sample = dict(mbr1=(1, uint8), mbr2=(2, uint16))

    def test_pack(self):
        assert items.pack(self.sample) == self.expected_bytes

    def test_pack_and_dump(self):
        memory, dump = items.pack_and_dump(self.sample)

        assert memory == self.expected_bytes
        assert str(dump) == self.expected_dump

    def test_bad_item(self):
        sample = dict(mbr1=(1, uint8), mbr2=2)

        with pytest.raises(PackError) as trap:
            items.pack(sample)

        expected = Baseline(
            """
            +--------+----------+-------+-------+---------------+
            | Offset | Access   | Value | Bytes | Format        |
            +--------+----------+-------+-------+---------------+
            |        |          |       |       | items         |
            | 0      | ['mbr1'] | 1     | 01    | uint8         |
            |        | ['mbr2'] | 2     |       | int (invalid) |
            +--------+----------+-------+-------+---------------+

            TypeError occurred during pack operation:

            no format specified, value must be a packable data type or (value,
            fmt) pairing (or dict/list of them)
            """
        )

        assert wrap_message(trap.value) == expected
        assert isinstance(trap.value.__context__, TypeError)



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Array to bytes and bytes to array transform."""

from typing import Any, Generator, List, Optional, Sequence, Tuple, Type, Union

from .data import Data, DataMeta
from .dump import Record
from .int import IntX
from .transform import Transform

Dims = Sequence[Optional[int]]
DimsArg = Union[Dims, "ArrayX", IntX]
Format = Union[Type[Data], Transform]

GREEDY_DIMS: Dims = (None,)


def iterate_dims(array: Sequence[Any], ndims: int) -> Generator[int, None, None]:
    """Determine dimensions from multi-dimensional array value."""
    for _ in range(ndims):
        yield len(array)
        array = array[0]


class ArrayX(Transform):

    """Array to bytes and bytes to array transform."""

    __ndims__: int
    __df__: Tuple[DimsArg, Format]

    def __init__(self, name: str, fmt: Format, dims: Optional[DimsArg] = None) -> None:
        super().__init__(name)

        assert isinstance(fmt, (DataMeta, Transform))

        locked_dims: DimsArg
        ndims = 0

        if dims is None:
            locked_dims = GREEDY_DIMS

        elif isinstance(dims, IntX):
            locked_dims = dims

        elif isinstance(dims, ArrayX):
            locked_dims = dims
            if not dims.__ndims__:
                raise TypeError(
                    "invalid ArrayX 'dims', must be single dim and non-greedy"
                )

        else:
            locked_dims = tuple(None if d is None else int(d) for d in dims)
            assert all(True if d is None else d > 0 for d in locked_dims)
            if (len(locked_dims) == 1) and locked_dims[0] is not None:
                ndims = locked_dims[0]

        nbytes: Optional[int] = fmt.__nbytes__

        if nbytes is not None:
            if isinstance(locked_dims, tuple):
                for dim in locked_dims:
                    if dim is None:
                        nbytes = None
                        break

                    nbytes *= dim

            else:
                nbytes = None

        self.__df__ = locked_dims, fmt
        self.__ndims__ = ndims
        self.__nbytes__ = nbytes

    @property
    def fmt(self) -> Format:
        """Array element format."""
        return self.__df__[1]

    @property
    def dims(self) -> DimsArg:
        """Array dimensions."""
        return self.__df__[0]

    def __unpack__(
        self,
        buffer: bytes,
        offset: int,
        dump: Optional[Record] = None,
        dims: Optional[Dims] = None,
    ) -> Tuple[List[Any], int]:
        # pylint: disable=arguments-differ,too-many-branches
        if dump is not None:
            return ArrayX.__unpack_and_dump__(self, buffer, offset, dump, dims)

        x_dims, fmt = self.__df__

        if dims is None:
            if isinstance(x_dims, tuple):
                dims = x_dims
                if None in dims[1:]:
                    raise TypeError(
                        "array unpack does not support greedy secondary dimensions"
                    )

            elif isinstance(x_dims, IntX):
                dim, offset = x_dims.__unpack__(buffer, offset, dump)
                dims = (dim,)

            elif isinstance(x_dims, ArrayX):
                dims, offset = x_dims.__unpack__(buffer, offset, dump)

            else:  # pragma: no cover
                raise RuntimeError(f"invalid 'dims' {x_dims!r}")

        array: List[Any] = []
        append = array.append

        this_dim, *item_dims = dims

        if this_dim is None:
            end = len(buffer)
            while offset < end:
                item, offset = fmt.__unpack__(buffer, offset, dump)
                append(item)

        elif item_dims:
            for _ in range(this_dim):
                item, offset = ArrayX.__unpack__(self, buffer, offset, dump, item_dims)
                append(item)
        else:
            for _ in range(this_dim):
                item, offset = fmt.__unpack__(buffer, offset, dump)
                append(item)

        return array, offset

    def __unpack_and_dump__(
        self,
        buffer: bytes,
        offset: int,
        dump: Record,
        dims: Optional[Dims] = None,
    ) -> Tuple[List[Any], int]:
        # pylint: disable=arguments-differ,too-many-locals,too-many-branches
        x_dims, fmt = self.__df__

        if dims is None:
            if isinstance(x_dims, tuple):
                dims = x_dims
                if None in dims[1:]:
                    raise TypeError(
                        "array unpack does not support greedy secondary dimensions"
                    )

            elif isinstance(x_dims, IntX):
                dim, offset = x_dims.__unpack__(
                    buffer, offset, dump.add_record(access="len()", fmt=x_dims)
                )
                dims = (dim,)

            elif isinstance(x_dims, ArrayX):
                dims, offset = x_dims.__unpack__(
                    buffer, offset, dump.add_record(access="--dims--", fmt=x_dims)
                )

            else:  # pragma: no cover
                raise RuntimeError(f"invalid 'dims' {x_dims!r}")

        array: List[Any] = []
        append = array.append

        this_dim, *item_dims = dims

        if this_dim is None:
            end = len(buffer)
            i = 0
            fmt_name = fmt.__name__
            while offset < end:
                item, offset = fmt.__unpack__(
                    buffer, offset, dump.add_record(access=f"[{i}]", fmt=fmt_name)
                )
                append(item)
                i += 1

        elif item_dims:
            for i in range(this_dim):
                item, offset = ArrayX.__unpack_and_dump__(
                    self, buffer, offset, dump.add_record(access=f"[{i}]"), item_dims
                )
                append(item)
        else:
            fmt_name = fmt.__name__
            for i in range(this_dim):
                item, offset = fmt.__unpack__(
                    buffer, offset, dump.add_record(access=f"[{i}]", fmt=fmt_name)
                )
                append(item)

        if not array:
            dump.value = []

        return array, offset

    def __pack__(
        self,
        value: Sequence[Any],
        pieces: List[bytes],
        dump: Optional[Record] = None,
        dims: Optional[Dims] = None,
    ) -> None:
        # pylint: disable=arguments-differ,too-many-branches
        if dump is not None:
            ArrayX.__pack_and_dump__(self, value, pieces, dump, dims)

        else:
            x_dims, fmt = self.__df__

            if dims is None:
                if isinstance(x_dims, tuple):
                    this_dim, *item_dims = x_dims

                elif isinstance(x_dims, IntX):
                    this_dim, item_dims = len(value), []
                    x_dims.__pack__(this_dim, pieces, dump)

                elif isinstance(x_dims, ArrayX):
                    dims = tuple(iterate_dims(value, x_dims.__ndims__))
                    this_dim, *item_dims = dims
                    x_dims.__pack__(dims, pieces, dump)

                else:  # pragma: no cover
                    raise RuntimeError(f"invalid 'dims' {x_dims!r}")

            else:
                this_dim, *item_dims = dims

            try:
                actual_length = len(value)
            except TypeError:
                raise TypeError("invalid array value, retrying with dump") from None

            if this_dim is not None and actual_length != this_dim:
                raise TypeError("invalid array value, retrying with dump")

            if item_dims:
                for item in value:
                    ArrayX.__pack__(self, item, pieces, dump, item_dims)
            else:
                for item in value:
                    fmt.__pack__(item, pieces, dump)

    def __pack_and_dump__(
        self,
        value: Sequence[Any],
        pieces: List[bytes],
        dump: Record,
        dims: Optional[Dims] = None,
    ) -> None:
        # pylint: disable=arguments-differ,too-many-branches
        x_dims, fmt = self.__df__

        if dims is None:
            if isinstance(x_dims, tuple):
                this_dim, *item_dims = x_dims

            elif isinstance(x_dims, IntX):
                this_dim, item_dims = len(value), []
                x_dims.__pack__(
                    this_dim, pieces, dump.add_record(access="len()", fmt=x_dims)
                )

            elif isinstance(x_dims, ArrayX):
                dims = tuple(iterate_dims(value, x_dims.__ndims__))
                this_dim, *item_dims = dims
                x_dims.__pack__(
                    dims, pieces, dump.add_record(access="--dims--", fmt=x_dims)
                )

            else:  # pragma: no cover
                raise RuntimeError(f"invalid 'dims' {x_dims!r}")

        else:
            this_dim, *item_dims = dims

        try:
            actual_length = len(value)
        except TypeError:
            dump.value = value
            raise TypeError(
                f"invalid value, expected iterable of "
                f'{"any " if this_dim is None else ""}'
                f'length{"" if this_dim is None else " " + str(this_dim)}'
                f", got non-iterable"
            ) from None

        if not actual_length:
            dump.value = []

        if this_dim is None:
            this_dim = actual_length

        if item_dims:
            for i, item in zip(range(this_dim), value):
                ArrayX.__pack_and_dump__(
                    self, item, pieces, dump.add_record(access=f"[{i}]"), item_dims
                )
        else:
            fmt_name = fmt.__name__

            for i, item in zip(range(this_dim), value):
                fmt.__pack__(
                    item, pieces, dump.add_record(access=f"[{i}]", fmt=fmt_name)
                )

        if actual_length != this_dim:
            for i in range(actual_length, this_dim):
                dump.add_record(access=f"[{i}]", value="<missing>")

            for i, item in zip(range(this_dim, actual_length), value[this_dim:]):
                dump.add_record(
                    access=f"[{i}] <extra>", value=item, separate=(i == this_dim)
                )
            raise TypeError(
                f"invalid value, expected iterable of "
                f"{this_dim} length, got iterable of length {actual_length}"
            )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Buffer corner cases."""

import pytest
from baseline import Baseline

from plum.conformance import wrap_message
from plum.buffer import Buffer
from plum.exceptions import UnpackError
from plum.littleendian import uint16


class TestUnpackErrors:

    expected_message = Baseline(
        """
        +--------+----------------------+-------+--------+
        | Offset | Value                | Bytes | Format |
        +--------+----------------------+-------+--------+
        | 0      | <insufficient bytes> | 01    | uint16 |
        +--------+----------------------+-------+--------+

        InsufficientMemoryError occurred during unpack operation:

        1 too few bytes to unpack 'uint16' (2 needed, only 1 available)
        """
    )

    def test_unpack(self):

        with pytest.raises(UnpackError) as trap:
            with Buffer(b"\x01") as buffer:
                buffer.unpack(uint16)

        assert wrap_message(trap.value) == self.expected_message

    def test_unpack_and_dump(self):
        with pytest.raises(UnpackError) as trap:
            with Buffer(b"\x01") as buffer:
                buffer.unpack_and_dump(uint16)

        assert wrap_message(trap.value) == self.expected_message



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test integer enumeration transform properties."""

from plum.enum import EnumX

from sample_enum import Register


class TestDefault:

    """Test with as many left to default as possible."""

    register = EnumX("register", Register)

    def test_name(self):
        assert self.register.name == "register"

    def test_enum(self):
        assert self.register.enum is Register

    def test_nbytes(self):
        assert self.register.nbytes == 1

    def test_byteorder(self):
        assert self.register.byteorder == "little"

    def test_signed(self):
        assert self.register.signed is False

    def test_strict(self):
        assert self.register.strict is True


class TestPositional:

    """Test explicitly defined with positional argument."""

    register = EnumX("register", Register, 2, "big", signed=True, strict=False)

    def test_name(self):
        assert self.register.name == "register"

    def test_enum(self):
        assert self.register.enum is Register

    def test_nbytes(self):
        assert self.register.nbytes == 2

    def test_byteorder(self):
        assert self.register.byteorder == "big"

    def test_signed(self):
        assert self.register.signed is True

    def test_strict(self):
        assert self.register.strict is False


class TestKeyword(TestPositional):

    """Test explicitly defined with keyword argument."""

    register = EnumX(
        name="register",
        enum=Register,
        nbytes=2,
        byteorder="big",
        signed=True,
        strict=False,
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""None to no bytes and no bytes to None transform."""

from typing import List, Optional, Tuple

from .dump import Record
from .transform import Transform


class NoneX(Transform):

    """None to no bytes and no bytes to None transform."""

    __nbytes__ = 0

    def __pack__(
        self, value: None, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        if dump is not None:
            dump.value = repr(value)

        if value is not None:
            raise TypeError("value must be 'None'")

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[None, int]:
        if dump is not None:
            dump.value = "None"

        return None, offset



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test miscellaneous features of AttrDict and AttrDictX transform."""

import pytest

from baseline import Baseline

from plum.attrdict import AttrDict, AttrDictX
from plum.array import ArrayX
from plum.conformance import wrap_message
from plum.exceptions import SizeError
from plum.littleendian import uint8, uint16


class TestAttrDictX:

    """Test AttrDictX features."""

    def test_repr_default_name(self):
        expected_repr = Baseline(
            """
            <transform 'AttrDict'>
            """
        )

        assert repr(AttrDictX(dict(m1=uint8, m2=uint16))) == expected_repr

    def test_repr_explicit_name(self):
        expected_repr = Baseline(
            """
            <transform 'EmptyDict'>
            """
        )

        assert (
            repr(AttrDictX(dict(m1=uint8, m2=uint16), name="EmptyDict"))
            == expected_repr
        )

    def test_size_error(self):
        """Verify no size when a member has no size."""
        expected_message = Baseline(
            """
            'AttrDict' transform sizes vary
            """
        )

        xform = AttrDictX({"m1": ArrayX(name="list", fmt=uint8), "m2": uint8})

        with pytest.raises(SizeError) as trap:
            xform.nbytes  # pylint: disable=pointless-statement

        assert wrap_message(trap.value) == expected_message


class TestAttrDict:

    """Test AttrDict properties."""

    def test_attribute_access(self):
        dct = AttrDict(m1=1)
        assert dct["m1"] == 1
        assert dct.m1 == 1  # pylint: disable=no-member

    def test_repr(self):
        expected_repr = Baseline(
            """
            AttrDict(m1=1, m2=2)
            """
        )

        assert repr(AttrDict(m1=1, m2=2)) == expected_repr



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2022 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Enumeration to bytes and bytes to enumeration transform."""

from enum import IntEnum
from typing import List, Optional, Tuple, Type

from ._getbytes import getbytes
from ._typing import ByteOrderHint
from .dump import Record
from .exceptions import InsufficientMemoryError
from .transform import Transform


class EnumX(Transform):

    """Enumeration to bytes and bytes to enumeration transform."""

    __nbytes__: int
    __bdnss__: Tuple[ByteOrderHint, Type[IntEnum], int, bool, bool]

    def __init__(
        self,
        name: str,
        enum: Type[IntEnum],
        nbytes: int = 1,
        byteorder: ByteOrderHint = "little",
        *,
        signed: bool = False,
        strict: bool = True,
    ) -> None:
        super().__init__(name)

        assert nbytes > 0
        assert byteorder in {"big", "little"}

        self.__nbytes__ = nbytes
        self.__benss__ = byteorder, enum, nbytes, bool(signed), bool(strict)

    @property
    def byteorder(self) -> ByteOrderHint:
        """Byte order ("little" or "big")."""
        byteorder, _enum, _nbytes, _signed, _strict = self.__benss__
        return byteorder

    @property
    def enum(self) -> Type[IntEnum]:
        """Enumeration."""
        _byteorder, enum, _nbytes, _signed, _strict = self.__benss__
        return enum

    @property
    def signed(self) -> bool:
        """Signed integer."""
        _byteorder, _enum, _nbytes, signed, _strict = self.__benss__
        return signed

    @property
    def strict(self) -> bool:
        """Values to pack/unpack must be an enumeration member."""
        _byteorder, _enum, _nbytes, _signed, strict = self.__benss__
        return strict

    def __pack__(
        self, value: int, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        byteorder, enum, nbytes, signed, strict = self.__benss__

        try:
            try:
                value = enum(value)
            except ValueError:
                if strict:
                    raise

            piece = int.to_bytes(value, nbytes, byteorder, signed=signed)
        except Exception:
            if dump is not None:
                # use repr in case str or something that otherwise looks like an int
                dump.value = repr(value)
            raise

        if dump is not None:
            dump.value = value
            dump.memory = piece

        pieces.append(piece)

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[int, int]:
        byteorder, enum, nbytes, signed, strict = self.__benss__

        if dump is None:
            end = offset + nbytes

            if len(buffer) < end:
                raise InsufficientMemoryError("too few bytes to unpack")

            value = int.from_bytes(buffer[offset:end], byteorder, signed=signed)

            try:
                return enum(value), end
            except ValueError:
                if strict:
                    raise
        else:
            piece, end = getbytes(buffer, offset, dump, nbytes)

            value = int.from_bytes(piece, byteorder, signed=signed)

            try:
                value = enum(value)
            except ValueError:
                if strict:
                    raise
            finally:  # pragma: no cover (coverage tool anomaly)
                dump.value = value

        return value, end

    def __call__(self, value: int) -> int:
        _byteorder, enum, _nbytes, _signed, strict = self.__benss__

        if strict:
            return enum(value)

        try:
            return enum(value)
        except ValueError:
            return value



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure nesting ability."""

from baseline import Baseline

from plum.conformance import Case, CaseData
from plum.littleendian import uint8
from plum.structure import Structure, member


class Nested(Structure):

    """Sample structure (for nesting inside another structure)."""

    inner: int = member(fmt=uint8)


class Outer(Structure):
    """Sample structure containing a nested structure."""

    outer: Nested = member(fmt=Nested)


class TestNested(Case):

    data = CaseData(
        fmt=Outer,
        bindata=b"\x00",
        nbytes=1,
        values=(
            Outer(outer=Nested(inner=0)),
            [[0]],
        ),
        dump=Baseline(
            """
            +--------+---------+-------+-------+--------+
            | Offset | Access  | Value | Bytes | Format |
            +--------+---------+-------+-------+--------+
            |        |         |       |       | Outer  |
            |        | outer   |       |       | Nested |
            | 0      |   inner | 0     | 00    | uint8  |
            +--------+---------+-------+-------+--------+
            """
        ),
        excess=Baseline(
            """
            +--------+---------+----------------+-------+--------+
            | Offset | Access  | Value          | Bytes | Format |
            +--------+---------+----------------+-------+--------+
            |        |         |                |       | Outer  |
            |        | outer   |                |       | Nested |
            | 0      |   inner | 0              | 00    | uint8  |
            +--------+---------+----------------+-------+--------+
            | 1      |         | <excess bytes> | 99    |        |
            +--------+---------+----------------+-------+--------+

            ExcessMemoryError occurred during unpack operation:

            1 unconsumed bytes
            """
        ),
        shortage=Baseline(
            """
            +--------+---------+----------------------+-------+--------+
            | Offset | Access  | Value                | Bytes | Format |
            +--------+---------+----------------------+-------+--------+
            |        |         |                      |       | Outer  |
            |        | outer   |                      |       | Nested |
            |        |   inner | <insufficient bytes> |       | uint8  |
            +--------+---------+----------------------+-------+--------+

            InsufficientMemoryError occurred during unpack operation:

            1 too few bytes to unpack 'uint8' (1 needed, only 0 available)
            """
        ),
        implementation=Baseline(
            """
            def __init__(self, *, outer: Nested) -> None:
                self[:] = (outer, )

            @outer.getter
            def outer(self) -> Nested:
                return self[0]

            @outer.setter
            def outer(self, value: Nested) -> None:
                self[0] = value

            @classmethod
            def __pack__(cls, value, pieces: List[bytes], dump: Optional[Record] = None) -> None:
                if isinstance(value, dict):
                    value = cls._make_structure_from_dict(value)

                (m_outer, ) = value

                if dump is None:
                    Nested.__pack__(m_outer, pieces, dump)

                else:
                    outer_dump = dump.add_record(access="outer", fmt=Nested)
                    Nested.__pack__(m_outer, pieces, outer_dump)

            @classmethod
            def __unpack__(cls, buffer: bytes, offset: int, dump: Optional[Record] = None) -> Tuple["Outer", int]:
                structure = list.__new__(cls)

                if dump is None:
                    m_outer, offset = Nested.__unpack__(buffer, offset, dump)

                else:
                    outer_dump = dump.add_record(access="outer", fmt=Nested)
                    m_outer, offset = Nested.__unpack__(buffer, offset, outer_dump)

                structure[:] = (m_outer, )

                return structure, offset

            def __eq__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__eq__(self, other)

            def __ne__(self, other: Any) -> bool:
                if isinstance(other, dict):
                    other = self._make_structure_from_dict(other)
                return list.__ne__(self, other)

            def __repr__(self) -> str:
                try:
                    return f"{type(self).__name__}(outer={self.outer!r})"
                except Exception:
                    return f"{type(self).__name__}()"

            """
        ),
    )



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Sized object to bytes and bytes to sized object transform."""

from typing import Any, List, Optional, Tuple, Type, Union

from ._getbytes import getbytes
from .int import IntX
from .data import Data
from .dump import Record
from .exceptions import ExcessMemoryError, InsufficientMemoryError
from .transform import Transform

Format = Union[Type[Data], Transform]


class SizedX(Transform):

    """Sized object to bytes and bytes to sized object transform.

    :param fmt: object format
    :param size_fmt: size format
    :param int ratio: number of bytes per increment of size member
    :param int offset: difference in size (in bytes)
    :param size_access: dump access description for size

    """

    __forss__: Tuple[Format, int, float, str, IntX]

    def __init__(
        self,
        name: str,
        *,
        fmt: Format,
        size_fmt: IntX,
        ratio: float = 1,
        offset: int = 0,
        size_access: str = "--size--",
    ) -> None:
        super().__init__(name)

        self.__forss__ = fmt, offset, ratio, size_access, size_fmt

    @property
    def fmt(self) -> Format:
        """Sized object format."""
        return self.__forss__[0]

    @property
    def offset(self) -> int:
        """Difference in size (in bytes)."""
        return self.__forss__[1]

    @property
    def ratio(self) -> float:
        """Number of bytes per increment of size."""
        return self.__forss__[2]

    @property
    def size_access(self) -> str:
        """Description to list in access column for size (e.g. "--size--")."""
        return self.__forss__[3]

    @property
    def size_fmt(self) -> IntX:
        """Size format."""
        return self.__forss__[4]

    def __pack__(
        self, value: Any, pieces: List[bytes], dump: Optional[Record] = None
    ) -> None:
        fmt, offset, ratio, size_access, size_fmt = self.__forss__

        if dump is None:
            size_record = None
            object_record = None
        else:
            size_record = dump.add_record(access=size_access, fmt=size_fmt)
            object_record = dump.add_record(fmt=fmt)

        object_pieces: List[bytes] = []

        try:
            fmt.__pack__(value, object_pieces, object_record)
        finally:
            if dump is not None and object_record is not None:
                dump += object_record[:]
                object_record[:] = []

        size = int(len(b"".join(object_pieces)) // ratio) + offset

        size_fmt.__pack__(size, pieces, size_record)

        pieces += object_pieces

    def __unpack__(
        self, buffer: bytes, offset: int, dump: Optional[Record] = None
    ) -> Tuple[Any, int]:
        # pylint: disable=too-many-locals
        fmt, size_offset, ratio, size_access, size_fmt = self.__forss__

        if dump is None:
            size_record = None
            object_record = None
        else:
            size_record = dump.add_record(access=size_access, fmt=size_fmt)
            object_record = dump.add_record(fmt=fmt)

        size, offset = size_fmt.__unpack__(buffer, offset, size_record)

        nbytes = int((size - size_offset) * ratio)
        end = offset + nbytes

        object_buffer = buffer[:end]

        try:
            if len(object_buffer) < end:
                if object_record is None:
                    # cause redo with dump to use getbytes() to make dump informative
                    raise InsufficientMemoryError("not enough buffer bytes")

                getbytes(buffer, offset, object_record, nbytes)

            value, offset = fmt.__unpack__(object_buffer, offset, object_record)

            extra_bytes = object_buffer[offset:]

            if extra_bytes:
                if object_record is not None:
                    for i in range(0, len(extra_bytes), 16):
                        if i:
                            object_record.add_record(memory=extra_bytes[i : i + 16])
                        else:
                            object_record.add_record(
                                separate=True,
                                value="<excess bytes>",
                                memory=extra_bytes[i : i + 16],
                            )

                raise ExcessMemoryError(extra_bytes)

        finally:
            if dump is not None and object_record is not None:  # pragma: no cover
                dump += object_record[:]
                object_record[:] = []

        return value, offset



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test float view to bytes buffer operations."""

from math import ceil, floor, trunc

import pytest
from baseline import Baseline

from plum.float import FloatX
from plum.utilities import pack

float32 = FloatX("float32", byteorder="big", nbytes=4)

# FUTURE: Rounding precision constant
# FUTURE: Clean up byte arrays


# pylint: disable=redefined-outer-name
@pytest.fixture()
def float32_view():
    """Sample float32 plum view."""
    buffer = b"\xff\x41\x45\x87\xe7\xff"
    yield float32.view(buffer, offset=1)  # float value of approx. 12.3456789


class TestFloatView:

    """Test operations float view to bytes buffer."""

    def test_arithmetic_assignments(self):
        """Test augmented arithmetic assignment operations."""
        buffer = bytearray([0xFF, 0x41, 0x45, 0x87, 0xE7, 0xFF])
        float32_view = float32.view(
            buffer, offset=1
        )  # float value of approx. 12.3456789

        # __iadd__
        float32_view += 1.234
        assert round(float32_view, 4) == round(13.579679489135742, 4)
        assert buffer == bytearray(b"\xffAYF^\xff")

        # __isub__
        float32_view -= 9.87654321
        assert round(float32_view, 4) == round(3.7031362791357427, 4)
        assert buffer == bytearray(b"\xff@m\x00/\xff")

        # __imul__
        float32_view *= 3.14159
        assert round(float32_view, 4) == round(11.633735656738281, 4)
        assert buffer == bytearray(b"\xffA:#\xc8\xff")

        # __imod__
        float32_view %= 2
        assert round(float32_view, 4) == round(1.6337356567382812, 4)
        assert buffer == bytearray(b"\xff?\xd1\x1e@\xff")

        # __ipow__
        float32_view **= 4
        assert round(float32_view, 4) == round(7.124053001403809, 4)
        assert buffer == bytearray(b"\xff@\xe3\xf8>\xff")

    def test_comparisons(self, float32_view):
        """Test comparison operations."""
        # pylint: disable=misplaced-comparison-constant

        # __lt__
        assert float32_view < 13.0
        assert not float32_view < 12.0

        # __le__
        assert float32_view <= 12.5
        assert not float32_view <= 12.25

        # __eq__
        assert float32_view == 12.34567928314209
        assert 12.34567928314209 == float32_view
        assert not float32_view == 98.67
        assert not 98.67 == float32_view

        # __ne__
        assert float32_view != 12.0
        assert 12.0 != float32_view

        # __gt__
        assert float32_view > 12.0
        assert not float32_view > 13.0

        # __ge__
        assert float32_view >= 12.25
        assert float32_view >= 12
        assert not float32_view >= 12.5
        assert not float32_view >= 13

    def test_numeric(self, float32_view):
        """Test comparison operations."""
        # __add__ and __radd__
        assert round(float32_view + 2, 4) == round(14.3456789, 4)
        assert round(2 + float32_view, 4) == round(14.3456789, 4)

        # __sub__ and __rsub__
        assert round(float32_view - 2, 4) == round(10.3456789, 4)
        assert round(20 - float32_view, 4) == round(7.6543211, 4)

        # __mul__ and __rmul__
        assert round(float32_view * 2, 4) == round(24.69135856628418, 4)
        assert round(2 * float32_view, 4) == round(24.69135856628418, 4)

        # __truediv__ and __rtruediv__
        assert round(float32_view / 3, 4) == round(4.11522642771403, 4)
        assert round(50 / float32_view, 4) == round(4.049999911165239, 4)

        # __mod__ and __rmod__
        assert round(float32_view % 5, 4) == round(2.34567928314209, 4)
        assert round(25.0 % float32_view, 4) == round(0.3086414337158203, 4)

        # __divmod__ and __rdivmod__
        assert divmod(float32_view, 3) == (4.0, 0.34567928314208984)
        assert divmod(24, float32_view) == (1.0, 11.65432071685791)

        # __pow__ and __rpow__
        assert round(float32_view**2, 4) == round(152.41579696220378, 4)
        assert round(2.0**float32_view, 4) == round(5204.988581065942, 4)

        # __ceil__ and __floor__
        assert ceil(float32_view) == 13
        assert floor(float32_view) == 12

        # __floordiv__ and __rfloordiv__
        assert float32_view // 2 == 6.0
        assert 30 // float32_view == 2.0

        # __round__ and __trunc__
        assert round(float32_view, 0) == 12
        assert trunc(float32_view) == 12

    def test_unary(self, float32_view):
        """Test unary operations."""
        assert round(-float32_view, 4) == round(-12.3456789, 4)  # __neg__
        assert round(+float32_view, 4) == round(12.3456789, 4)  # __pos__
        assert round(abs(float32_view), 4) == round(12.3456789, 4)  # __abs__
        assert int(float32_view) == 12  # __int__

    def test_pack_pos_argument(self, float32_view):
        """Test pack utility supports view as positional argument."""
        assert pack(float32_view, float32) == bytearray(b"AE\x87\xe7")

    def test_pack_kw_argument(self, float32_view):
        """Test pack utility supports view as keyword argument."""
        assert pack({"x": float32_view}, {"x": float32}) == bytearray(b"AE\x87\xe7")

    def test_unpack(self):
        """Test unpack() returns unpacked value."""
        buffer = bytearray(b"\xff\x41\x45\x87\xe7\xff")
        value = float32.view(buffer, offset=1).unpack()
        assert value == 12.34567928314209
        # changing buffer does not effect previously returned value
        buffer[0:6] = [0] * 6
        assert value == 12.34567928314209

    def test_set(self):
        """Test set()."""
        buffer = bytearray(6)
        float_view = float32.view(buffer, offset=1)
        assert float_view == 0.0
        float_view.set(12.34567928314209)
        assert float_view == 12.34567928314209
        assert buffer == bytearray([0x00, 0x41, 0x45, 0x87, 0xE7, 0x00])

    def test_representations(self, float32_view):
        """Test representations."""
        expected_dump = Baseline(
            """
                    +--------+-------------------+-------------+---------+
                    | Offset | Value             | Bytes       | Format  |
                    +--------+-------------------+-------------+---------+
                    | 1      | 12.34567928314209 | 41 45 87 e7 | float32 |
                    +--------+-------------------+-------------+---------+
                    """
        )

        assert str(float32_view) == "12.34567928314209"
        assert repr(float32_view) == "<view at 0x1: float32(12.34567928314209)>"
        assert str(float32_view.dump) == expected_dump



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Pack/unpack bytes summary."""

from typing import TYPE_CHECKING, Any, Generator, List, Tuple, Type, Union

Format = Union[str, "transform.Transform", Type["data.Data"]]


class Base(list):

    """Dump/Record Base class."""

    def add_record(
        self,
        access: str = "",
        value: Any = "",
        bits: Union[Tuple[int, int], None] = None,
        memory: bytes = b"",
        fmt: Format = "",
        separate: bool = False,
    ) -> "Record":
        """Add child record.

        :param access: access description
        :param value: representation of value associated with bytes
        :param bits: start, end bit offset (None -> not a bit field)
        :param memory: bytes
        :param fmt: bytes format
        :param separate: add separator before row

        """
        # pylint: disable=too-many-arguments
        record = Record(access, value, bits, memory, fmt, separate)
        self.append(record)
        return record

    def add_extra_bytes(self, access: str, memory: bytes):
        """Add records listing bytes without access/value descriptions.

        :param access: access description
        :param memory: memory bytes

        """
        for i in range(0, len(memory), 16):
            self.add_record(access=access, memory=memory[i : i + 16])
            access = ""


class Record(Base):

    """Pack/unpack bytes summary.

    :param access: access description
    :param value: representation of value associated with bytes
    :param bits: start, end bit offset (None -> not a bit field)
    :param memory: memory bytes
    :param fmt: class name
    :param separate: add separator before row

    """

    def __init__(
        self,
        access: str = "",
        value: Any = "",
        bits: Union[Tuple[int, int], None] = None,
        memory: bytes = b"",  # pylint: disable=used-before-assignment ???
        fmt: Format = "",
        separate: bool = False,
    ) -> None:
        # pylint: disable=too-many-arguments
        super().__init__()
        self.access = access
        self._value = value
        self._bits = bits
        self.memory = memory
        self._fmt = fmt
        self.separate = separate

    def __repr__(self):  # pragma: no cover
        return (
            f"Record("
            f"bits={self.bits!r}, "
            f"access={self.access!r}, "
            f"value={self.value}, "
            f"memory={self.memory}, "
            f"fmt={self.fmt}, "
            f"records={list(self)})"
        )

    def iter_records(
        self, indent: int = 0
    ) -> Generator[Tuple[int, "Record"], None, None]:
        """Iterate dump records."""
        yield indent, self
        for record in self:
            yield from record.iter_records(indent + 1)

    @property
    def bits(self) -> str:
        """Bit field representation."""
        if self._bits is None:
            bits = ""
        else:
            lsb, size = self._bits

            if size == 1:
                bits = f"[{lsb}]"
            else:
                bits = f"[{lsb}:{lsb + size}]"

        return bits

    @property
    def value(self) -> Any:
        """Value representation."""
        # FUTURE - Any return really should be str but mypy complains
        #          leave this return self._value and leave application
        #          convert?
        return str(self._value)

    @value.setter
    def value(self, value: Any) -> None:
        """Value representation."""
        self._value = value

    @staticmethod
    def hack(fmt: Any) -> str:
        """Return class name if a class, otherwise convert to a str."""
        # FUTURE get rid of this once everyone commits to submitting strings
        # or we go back to accepting types and we get mypy figured out
        try:
            return fmt.__name__
        except AttributeError:
            return str(fmt)

    @property
    def fmt(self) -> str:
        """Entry format."""
        return self.hack(self._fmt)

    @fmt.setter
    def fmt(self, value: str) -> None:
        self._fmt = value

    @property
    def is_blank(self):
        """Has blank access, value, memory, and format fields."""
        return (
            all(x == "" for x in (self._fmt, self.access, self._value))
            and self.memory == b""
        )


class Dump(Base):

    """Pack/unpack bytes summary.

    :param offset: master byte offset

    """

    def __init__(self, offset: int = 0) -> None:
        super().__init__()
        self.offset: int = offset

    def __repr__(self):  # pragma: no cover
        return f"Dump({list(self)})"

    def __call__(self) -> None:
        print(self)

    def iter_records(self) -> Generator[Tuple[int, "Record"], None, None]:
        """Iterate dump records, yield (level, record) pairs."""
        for record in self:
            yield from record.iter_records()

    @staticmethod
    def _get_bytes_column_cells(records) -> List[str]:
        bitfield_size = 0
        cells = []
        for record in records:
            memory = record.memory
            if memory:
                cells.append(" ".join("{:02x}".format(c) for c in memory))
                bitfield_size = len(memory) * 8
            elif record._bits is None:  # pylint: disable=protected-access
                cells.append("")
            else:
                value = record._value  # pylint: disable=protected-access
                chars = ["."] * bitfield_size
                offset, size = record._bits  # pylint: disable=protected-access
                while size:
                    try:
                        chars[offset] = str(value & 1)
                    except (IndexError, TypeError):  # pragma: no cover
                        # sometimes records are half baked during a transform and didn't
                        # get a chance to complete before a problem occurred
                        break
                    value >>= 1
                    offset += 1
                    size -= 1
                chars.reverse()
                for i in range(bitfield_size - 8, 0, -8):
                    chars[i:i] = [" "]
                cells.append("".join(chars))
        return cells

    def _get_offset_column_cells(self, records) -> List[str]:
        bits = [record.bits for record in records]

        # make bit offset information uniform in length
        if any(bits):
            fmt = "{:%ds}" % max(len(bits_desc) for bits_desc in bits)
            bits = [fmt.format(bits_desc) for bits_desc in bits]

        # compute offset cells (with bit descriptions appended on)
        nbytes = sum(len(record.memory) for record in records)
        offset_template = "{:%dd}" % len(str(nbytes + self.offset))
        filler = " " * len(str(nbytes))
        consumed = 0
        offsets = []
        for record, bits_desc in zip(records, bits):
            if record.memory:
                byte_offset = offset_template.format(self.offset + consumed)
                consumed += len(record.memory)
            else:
                byte_offset = filler
            offsets.append(byte_offset + bits_desc)

        return offsets

    def _get_lines(self) -> Generator[str, None, None]:
        if len(self):  # pylint: disable=len-as-condition
            indents, records = zip(*self.iter_records())
        else:
            indents, records = (0,), (Record(),)

        accesses = [record.access.replace("(.)", "") for record in records]

        # look for things like this and get rid of excess indentation
        # +--------+-------------+-----------+-------+-----------------+
        # | Offset | Access      | Value     | Bytes | Format          |
        # +--------+-------------+-----------+-------+-----------------+
        # |        |             |           |       | CustomStructure |
        # | 0      |   [0] (.m1) | 258       | 02 01 | uint16          |
        # | 2      |   [1] (.m2) | <invalid> | 00    | CustomError     |
        # +--------+-------------+-----------+-------+-----------------+
        if all(indent or not access for indent, access in zip(indents, accesses)):
            indents = tuple(indent - 1 if indent else 0 for indent in indents)

        columns = [
            ["Offset"] + self._get_offset_column_cells(records),
            # 'Access' gets inserted here
            ["Value"] + [record.value for record in records],
            ["Bytes"] + self._get_bytes_column_cells(records),
            ["Format"] + [record.fmt for record in records],
        ]

        if any(accesses):
            columns[1:1] = [
                ["Access"]
                + ["  " * indent + access for indent, access in zip(indents, accesses)]
            ]

        column_widths = [max([len(cell) for cell in column]) for column in columns]

        border = "+{}+".format("+".join("-" * (n + 2) for n in column_widths))
        row_template = "|{}|".format("|".join(" {:%ds} " % n for n in column_widths))

        rows = list(zip(*columns))
        separators = [record.separate for record in records]
        separators[0] = False

        yield border
        yield row_template.format(*rows.pop(0))  # header names
        yield border
        for cells, separate in zip(rows, separators):
            if separate:
                yield border
            yield row_template.format(*cells)
        yield border

    def __str__(self) -> str:
        return "\n".join(self._get_lines())

    def trim_blank_record(self):
        """Remove first record and move its child records up if nothing to show."""
        first_record = self[0]
        if first_record.is_blank:
            self[:] = first_record[:]


if TYPE_CHECKING:  # pragma: no cover
    from . import data
    from . import transform



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Copyright 2021 Daniel Mark Gass, see __about__.py for license information.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
"""Test Structure type inheritance behaviors."""

from baseline import Baseline

from plum.littleendian import uint8
from plum.structure import Structure, member


class Base(Structure):

    m0 = member(fmt=uint8)
    m1 = member(fmt=uint8)


class TestAddMethod:

    """Inherit members, but just add a method."""

    class Struct(Base):
        @property
        def product(self):
            return self.m0 * self.m1

    def test_init_and_property(self):
        assert self.Struct(m0=2, m1=3).product == 6

    def test_round_trip(self):
        expected_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        |        |       |       | Struct |
            | 0      | m0     | 1     | 01    | uint8  |
            | 1      | m1     | 2     | 02    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        )
        original = self.Struct(m0=1, m1=2)
        buffer, pack_dump = original.ipack_and_dump()
        copy, unpack_dump = self.Struct.unpack_and_dump(buffer)
        assert str(pack_dump) == expected_dump
        assert str(unpack_dump) == expected_dump
        assert original == copy


class TestAddMembers:

    """Inherit members, but just add a method."""

    class Struct(Base):

        m2 = member(fmt=uint8)
        m3 = member(fmt=uint8)

    def test_round_trip(self):
        expected_dump = Baseline(
            """
            +--------+--------+-------+-------+--------+
            | Offset | Access | Value | Bytes | Format |
            +--------+--------+-------+-------+--------+
            |        |        |       |       | Struct |
            | 0      | m0     | 1     | 01    | uint8  |
            | 1      | m1     | 2     | 02    | uint8  |
            | 2      | m2     | 3     | 03    | uint8  |
            | 3      | m3     | 4     | 04    | uint8  |
            +--------+--------+-------+-------+--------+
            """
        )
        original = self.Struct(m0=1, m1=2, m2=3, m3=4)
        buffer, pack_dump = original.ipack_and_dump()
        copy, unpack_dump = self.Struct.unpack_and_dump(buffer)
        assert str(pack_dump) == expected_dump
        assert str(unpack_dump) == expected_dump
        assert original == copy
